@@ -21,7 +21,7 @@ Components.utils.import("resource://gre/modules/ctypes.jsm");
 
 const mozISpellCheckingEngine = Components.interfaces.mozISpellCheckingEngine;
 const CLASS_ID = Components.ID("89630d4c-c64d-11e0-83d8-00508d9f364f");
-const CLASS_NAME = "Finnish language spell-check";
+const CLASS_NAME = "Spell checker based on Voikko";
 const CONTRACT_ID = "@mozilla.org/mozvoikko2;1";
 
 const VOIKKO_OPT_IGNORE_DOT = 0;
@@ -29,12 +29,42 @@ const VOIKKO_OPT_IGNORE_NUMBERS = 1;
 const VOIKKO_OPT_IGNORE_UPPERCASE = 3;
 const VOIKKO_OPT_ACCEPT_MISSING_HYPHENS = 12;
 
+const OPTIONAL_DEPENDENCIES = {
+    "WINNT_x86-msvc" : ["libgcc_s_sjlj-1.dll", "libstdc++-6.dll", "zlib1.dll", "libarchive-13.dll", "libhfstospell-4.dll"],
+    "WINNT_x86_64-msvc" : ["libgcc_s_seh-1.dll", "libstdc++-6.dll", "zlib1.dll", "libarchive-13.dll", "libhfstospell-4.dll"],
+    "Darwin_x86_64-gcc3" : ["libtinyxml2.3.dylib", "libarchive.13.dylib", "libhfstospell.4.dylib"],
+    "Darwin_x86-gcc3" : ["libtinyxml2.3.dylib", "libarchive.13.dylib", "libhfstospell.4.dylib"]
+};
+
+function tryLoadLibrary(abi, lib_name) {
+    // Try to locate the library inside extension directory and
+    // replace value
+    var extension_dir = __LOCATION__.parent.parent;
+    var library_loc = extension_dir.clone();
+    library_loc.append("voikko");
+    library_loc.append(abi);
+    library_loc.append(lib_name);
+    if (library_loc.exists() && library_loc.isFile())
+    {
+        var handle = ctypes.open(library_loc.path);
+        aConsoleService.logStringMessage("MozVoikko2: loaded bundled library " + library_loc.path);
+        return handle;
+    }
+    else
+    {
+        var handle = ctypes.open(lib_name);
+        aConsoleService.logStringMessage("MozVoikko2: loaded system library " + library_loc.path);
+        return handle;
+    }
+}
+
 function LibVoikko()
 {
 }
 
 LibVoikko.prototype = {
     libvoikko: null,
+    optional_deps: [],
 
     data_loc: "",
 
@@ -46,6 +76,7 @@ LibVoikko.prototype = {
     fn_voikko_suggest_cstr: null,
     fn_voikko_free_cstr_array: null,
     fn_voikko_set_boolean_option: null,
+    fn_voikkoListSupportedSpellingLanguages: null,
 
     init : function()
     {
@@ -59,16 +90,16 @@ LibVoikko.prototype = {
         //==================================================================
         // Detect shared library name to load
         //==================================================================
-        if (abi == "Linux_x86-gcc3" || abi == "Linux_x86_64-gcc3")
+        if (abi.indexOf("Linux_") == 0)
         {
             lib_name = "libvoikko.so.1";
         }
-        else if (abi == "WINNT_x86-msvc")
+        else if (abi == "WINNT_x86-msvc" || abi == "WINNT_x86_64-msvc")
         {
             lib_name = "libvoikko-1.dll";
             this.call_abi = ctypes.winapi_abi;
         }
-        else if (abi == "Darwin_x86_64-gcc3" || abi == "Darwin_x86-gcc3")
+        else if (abi == "Darwin_x86_64-gcc3" || abi == "Darwin_x86-gcc3" || abi == "Darwin_ppc-gcc3")
         {
             lib_name = "libvoikko.1.dylib";
         }
@@ -77,36 +108,32 @@ LibVoikko.prototype = {
             throw "Unsupported ABI " + abi;
         }
 
-        // Try to locate libvoikko inside extension directory and
-        // replace value
-        var extension_dir = __LOCATION__.parent.parent;
-        var libvoikko_loc = extension_dir.clone();
-        libvoikko_loc.append("voikko");
-        libvoikko_loc.append(abi);
-        libvoikko_loc.append(lib_name);
-        if (libvoikko_loc.exists() && libvoikko_loc.isFile())
-        {
-            this.libvoikko = ctypes.open(libvoikko_loc.path);
-            aConsoleService.logStringMessage("MozVoikko2: loaded " + libvoikko_loc.path);
+        try {
+            this.libvoikko = tryLoadLibrary(abi, lib_name);
         }
-        else
-        {
-            this.libvoikko = ctypes.open(lib_name);
-            aConsoleService.logStringMessage("MozVoikko2: loaded system libvoikko");
-            aConsoleService.logStringMessage("MozVoikko2: libvoikko_loc = " + libvoikko_loc.path);
+        catch (err) {
+            aConsoleService.logStringMessage("MozVoikko2: Failed to load libvoikko (" + lib_name + "), maybe we need optional dependencies");
+            if (abi in OPTIONAL_DEPENDENCIES) {
+                for (var i = 0; i < OPTIONAL_DEPENDENCIES[abi].length; i++) {
+                    var dep = OPTIONAL_DEPENDENCIES[abi][i];
+                    try {
+                        var h = tryLoadLibrary(abi, dep);
+                        this.optional_deps.push(h);
+                    }
+                    catch (e2) {
+                        aConsoleService.logStringMessage("MozVoikko2: Failed to load optional dependency " + dep);
+                    }
+                }
+                this.libvoikko = tryLoadLibrary(abi, lib_name);
+            }
+            else {
+                throw err;
+            }
         }
 
-        var data_loc = extension_dir.clone();
+        var data_loc = __LOCATION__.parent.parent.clone();
         data_loc.append("voikko");
-        var data_loc_test = data_loc.clone();
-        data_loc_test.append("2");
-        data_loc_test.append("mor-standard");
-        data_loc_test.append("voikko-fi_FI.pro");
-        if (data_loc_test.exists() && data_loc_test.isFile())
-        {
-            this.data_loc = data_loc.path;
-            aConsoleService.logStringMessage("MozVoikko2: Found suomi-malaga data at " + this.data_loc);
-        }
+        this.data_loc = data_loc.path;
 
         /* Detect used libvoikko version and output a message to Javascript console. */
         var fn_voikkoGetVersion = this.libvoikko.declare(
@@ -178,6 +205,16 @@ LibVoikko.prototype = {
             ctypes.voidptr_t,
             ctypes.int,
             ctypes.int);
+
+        //
+        // char ** voikkoListSupportedSpellingLanguages(const char * path)
+        //
+        this.fn_voikkoListSupportedSpellingLanguages = this.libvoikko.declare(
+            "voikkoListSupportedSpellingLanguages",
+            this.call_abi,
+            ctypes.char.ptr.array(50).ptr,
+            ctypes.char.ptr);
+
     },
 
     finalize : function()
@@ -185,6 +222,9 @@ LibVoikko.prototype = {
         this.fn_voikko_init = null;
         this.fn_voikko_terminate = null;
         this.libvoikko.close();
+        while (this.optional_deps.length > 0) {
+            this.optional_deps.pop().close();
+        }
     }
 };
 
@@ -211,7 +251,6 @@ function VoikkoHandle(libvoikko)
 VoikkoHandle.prototype = {
     handle: null,
     libvoikko: null,
-    lang_code: null,
 
     open : function(libvoikko, lang_code)
     {
@@ -228,8 +267,12 @@ VoikkoHandle.prototype = {
 
         this.handle = this.libvoikko.fn_voikko_init(
             message_ptr,
-            "fi_FI",
+            lang_code,
             data_loc);
+        if (this.handle.isNull()) {
+            this.handle = null;
+            throw "Could not load dictionary that should exists: " + lang_code;
+        }
 
         this.libvoikko.fn_voikko_set_boolean_option(this.handle, VOIKKO_OPT_IGNORE_DOT, 1);
         this.libvoikko.fn_voikko_set_boolean_option(this.handle, VOIKKO_OPT_IGNORE_NUMBERS, 1);
@@ -239,7 +282,7 @@ VoikkoHandle.prototype = {
 
     finalize : function()
     {
-        if (handle)
+        if (this.handle)
         {
             this.libvoikko.fn_voikko_terminate(this.handle);
             this.handle = null;
@@ -250,16 +293,6 @@ VoikkoHandle.prototype = {
 
 function MozVoikko2()
 {
-    try
-    {
-        this.voikko_handle = new VoikkoHandle;
-        this.voikko_handle.open(libvoikko);
-    }
-    catch (err)
-    {
-        Components.utils.reportError(err);
-        throw err;
-    }
 }
 
 MozVoikko2.prototype = {
@@ -267,23 +300,44 @@ MozVoikko2.prototype = {
     classID: CLASS_ID,
     contractID: CONTRACT_ID,
 
-    data_loc : null,
     voikko_handle : null,
+    supportedDicts : null,
+    currentDict: null,
     mPersonalDictionary : null,
 
     QueryInterface: XPCOMUtils.generateQI([mozISpellCheckingEngine, Components.interfaces.nsISupport]),
 
     get dictionary()
     {
-        return "fi_FI";
+        return this.currentDict;
     },
 
     set dictionary(dict)
     {
-        if (dict != "fi_FI")
+        if (dict == this.currentDict)
+        {
+            return;
+        }
+        var availableDict = this.getSupportedDictionaryCodeInternal(dict);
+        if (availableDict == null)
         {
             throw "mozvoikko2: dictionary '" + dict + "' is not supported by this component (but may be supported by others)";
         }
+        try
+        {
+            if (this.voikko_handle)
+            {
+                this.voikko_handle.finalize();
+            }
+            this.voikko_handle = new VoikkoHandle;
+            this.voikko_handle.open(libvoikko, availableDict);
+            this.currentDict = dict;
+        }
+        catch (err)
+        {
+            Components.utils.reportError(err);
+            throw err;
+        }
     },
 
     get providesPersonalDictionary()
@@ -316,10 +370,53 @@ MozVoikko2.prototype = {
         this.mPersonalDictionary = mPersonalDictionary;
     },
 
+    getSupportedDictionariesInternal : function()
+    {
+        if (this.supportedDicts == null) {
+            var data_loc = null;
+            if (libvoikko.data_loc) {
+                data_loc = libvoikko.data_loc;
+            }
+            var cSpellingLangs = libvoikko.fn_voikkoListSupportedSpellingLanguages(data_loc);
+            if (!cSpellingLangs.isNull())
+            {
+                var spellingLangs = [];
+                for (i = 0; i < 200 && !cSpellingLangs.contents[i].isNull(); i++)
+                {
+                    spellingLangs.push(cSpellingLangs.contents[i].readString());
+                }
+                libvoikko.fn_voikko_free_cstr_array(cSpellingLangs);
+                this.supportedDicts = spellingLangs;
+            }
+        }
+    },
+
+    getSupportedDictionaryCodeInternal : function(dict)
+    {
+        this.getSupportedDictionariesInternal();
+        if (this.supportedDicts == null)
+        {
+            return;
+        }
+        if (this.supportedDicts.indexOf(dict) != -1)
+        {
+            return dict;
+        }
+        if (dict.indexOf("_") > -1)
+        {
+            var d = dict.substr(0, dict.indexOf("_"));
+            if (this.supportedDicts.indexOf(d) != -1)
+            {
+                return d;
+            }
+        }
+    },
+
     getDictionaryList : function(dicts, count)
     {
-        dicts.value = ["fi_FI"];
-        count.value = 1;
+        this.getSupportedDictionariesInternal();
+        dicts.value = this.supportedDicts;
+        count.value = this.supportedDicts.length;
     },
 
     check : function(word)
@@ -332,7 +429,7 @@ MozVoikko2.prototype = {
 
         if (result == 0 && this.mPersonalDictionary)
         {
-            return this.mPersonalDictionary.check(word, "fi_FI");
+            return this.mPersonalDictionary.check(word, this.currentDict);
         }
         else
         {
